from collections import defaultdict
from typing import Callable, Dict, Optional, TypeVar, Set, cast, Iterator
import functools
import dask

A = TypeVar("A")
B = TypeVar("B")
K = TypeVar("K")
U = TypeVar("U")


def map(self: Set[A], f: Callable[[A], B]) -> Set[B]:
    """
    Builds a new set by applying a function to all elements of this set.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new set.
    """
    return cast(Set[B], type(self)(f(x) for x in self))


def filter(self: Set[A], p: Callable[[A], bool]) -> Set[A]:
    """
    Selects all elements of this set which satisfy a predicate.

    Args:
        p: The predicate to satisfy.

    Returns:
        The filtered set.
    """
    return type(self)(x for x in self if p(x))


def filter_not(self: Set[A], p: Callable[[A], bool]) -> Set[A]:
    """
    Selects all elements of this set which do not satisfy a predicate.

    Args:
        p: The predicate to not satisfy.

    Returns:
        The filtered set.
    """
    return type(self)(x for x in self if not p(x))


def flat_map(self: Set[A], f: Callable[[A], Set[B]]) -> Set[B]:
    """
    Builds a new set by applying a function to all elements of this set and using the elements of the resulting collections.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new set.
    """
    return cast(Set[B], type(self)(y for x in self for y in f(x)))


def contains(self: Set[A], elem: A) -> bool:
    """
    Tests whether this set contains a given value as element.

    Args:
        elem: The element to look for.

    Returns:
        True if the set contains the element, False otherwise.
    """
    return elem in self


def foreach(self: Set[A], f: Callable[[A], U]) -> None:
    """
    Apply f to each element of the set for its side effects.

    Args:
        f: The function to apply to all elements for its side effects.
    """
    for x in self:
        f(x)


def group_by(self: Set[A], f: Callable[[A], K]) -> Dict[K, Set[A]]:
    """
    Partitions this set into a dict of sets according to some discriminator function.

    Args:
        f: The grouping function.

    Returns:
        A dictionary where elements are grouped according to the grouping function.
    """
    # frozenset does not have `add`
    d = defaultdict(set if isinstance(self, frozenset) else type(self))
    for x in self:
        k = f(x)
        d[k].add(x)
    return d


def is_empty(self: Set[A]) -> bool:
    """
    Tests whether the set is empty.

    Returns:
        True if the set is empty, False otherwise.
    """
    return len(self) == 0


def size(self: Set[A]) -> int:
    """
    Computes the size of this set.

    Returns:
        The size of the set.
    """
    return len(self)


def find(self: Set[A], p: Callable[[A], bool]) -> Optional[A]:
    """
    Finds the first element of the set satisfying a predicate, if any.

    Args:
        p: The predicate to satisfy.

    Returns:
        The first element satisfying the predicate, otherwise None.
    """
    for x in self:
        if p(x):
            return x
    return None


def fold_left(self: Set[A], z: B, op: Callable[[B, A], B]) -> B:
    """
    Applies a binary operator to a start value and all elements of this set, going left to right.

    Note: might return different results for different runs, unless the underlying collection type is ordered or the operator is associative and commutative.

    Args:
        z: The start value.
        op: The binary operation.

    Returns:
        The result of inserting op between consecutive elements of this set, going left to right with the start value z on the left:
        op(...op(z, x_1), x_2, ..., x_n)
        where x1, ..., xn are the elements of this set. Returns z if this set is empty.
    """
    acc = z
    for x in self:
        acc = op(acc, x)
    return acc


def fold_right(self: Set[A], z: B, op: Callable[[A, B], B]) -> B:
    """
    Applies a binary operator to all elements of this set and a start value, going right to left.

    Note: might return different results for different runs, unless the underlying collection type is ordered or the operator is associative and commutative.

    Args:
        z: The start value.
        op: The binary operation.

    Returns:
        The result of inserting op between consecutive elements of this set, going right to left with the start value z on the right:
        op(x_1, op(x_2, ... op(x_n, z)...))
        where x1, ..., xn are the elements of this set. Returns z if this set is empty.
    """

    acc = z
    for x in self:
        acc = op(x, acc)
    return acc


def forall(self: Set[A], p: Callable[[A], bool]) -> bool:
    """
    Tests whether a predicate holds for all elements of this set.

    Args:
        p: The predicate used to test elements.

    Returns:
        True if this set is empty or the given predicate p holds for all elements of this set, otherwise False.
    """
    for x in self:
        if not p(x):
            return False

    return True


def length(self: Set[A]) -> int:
    """
    Returns the length (number of elements) of the set. `size` is an alias for length.

    Returns:
        The length of the set
    """
    return len(self)


def to_iterator(self: Set[A]) -> Iterator[A]:
    """
    Converts this set to an iterator.

    Returns:
        An iterator
    """
    return (x for x in self)

# Parallel operations


def par_map(self: Set[A], f: Callable[[A], B]) -> Set[B]:
    """
    Builds a new set by applying in parallel a function to all elements of this set.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new set.
    """
    return cast(Set[B], type(self)((dask.compute(*(dask.delayed(f)(x) for x in self)))))


def par_filter(self: Set[A], p: Callable[[A], bool]) -> Set[A]:
    """
    Selects in parallel all elements of this set which satisfy a predicate.

    Args:
        p: The predicate to satisfy.

    Returns:
        The filtered set.
    """
    preds = dask.compute(*(dask.delayed(p)(x) for x in self))
    return type(self)(x for i, x in enumerate(self) if preds[i])


def par_filter_not(self: Set[A], p: Callable[[A], bool]) -> Set[A]:
    """
    Selects in parallel all elements of this set which do not satisfy a predicate.

    Args:
        p: The predicate to not satisfy.

    Returns:
        The filtered set.
    """
    preds = dask.compute(*(dask.delayed(p)(x) for x in self))
    return type(self)(x for i, x in enumerate(self) if not preds[i])


def par_flat_map(self: Set[A], f: Callable[[A], Set[B]]) -> Set[B]:
    """
    Builds a new set by applying in parallel a function to all elements of this set and using the elements of the resulting collections.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new set.
    """
    applications = dask.compute(*(dask.delayed(f)(x) for x in self))
    return cast(Set[B], type(self)(x for y in applications for x in y))


# Pure operations


def pure_map(self: Set[A], f: Callable[[A], B]) -> Set[B]:
    """
    Builds a new set by applying a function to all elements of this set using memoization to improve performance.

    WARNING: f must be a PURE function i.e., calling f on the same input must always lead to the same result!

    Type A must be hashable using `hash()` function.

    Args:
        f: The PURE function to apply to all elements.

    Returns:
        The new set.
    """
    f_cache = functools.cache(f)
    return cast(Set[B], type(self)(f_cache(x) for x in self))


def pure_flat_map(self: Set[A], f: Callable[[A], Set[B]]) -> Set[B]:
    """
    Builds a new set by applying a function to all elements of this set and using the elements of the resulting collections using memoization to improve performance.

    WARNING: f must be a PURE function i.e., calling f on the same input must always lead to the same result!

    Type A must be hashable using `hash()` function.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new set.
    """
    f_cache = functools.cache(f)
    return cast(Set[B], type(self)(y for x in self for y in f_cache(x)))


def pure_filter(self: Set[A], p: Callable[[A], bool]) -> Set[A]:
    """
    Selects all elements of this set which satisfy a predicate using memoization to improve performance.

    WARNING: p must be a PURE function i.e., calling p on the same input must always lead to the same result!

    Type A must be hashable using `hash()` function.

    Args:
        p: The predicate to satisfy.

    Returns:
        The filtered set.
    """
    p_cache = functools.cache(p)
    return type(self)(x for x in self if p_cache(x))


def pure_filter_not(self: Set[A], p: Callable[[A], bool]) -> Set[A]:
    """
    Selects all elements of this set which do not satisfy a predicate using memoization to improve performance.

    WARNING: p must be a PURE function i.e., calling p on the same input must always lead to the same result!

    Type A must be hashable using `hash()` function.


    Args:
        p: The predicate not to satisfy.

    Returns:
        The filtered set.
    """
    p_cache = functools.cache(p)
    return type(self)(x for x in self if not p_cache(x))


def lazy_map(self: Set[A], f: Callable[[A], B]) -> Iterator[B]:
    """
    Builds a new set by applying a function to all elements of this set, lazily.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new lazy set, as an iterator.
    """
    for x in self:
        yield f(x)


def lazy_filter(self: Set[A], p: Callable[[A], bool]) -> Iterator[A]:
    """
    Selects all elements of this set which satisfy a predicate, lazily.

    Args:
        p: The predicate to satisfy.

    Returns:
        The filtered lazy set, as an iterator.
    """
    for x in self:
        if p(x):
            yield x


def lazy_filter_not(self: Set[A], p: Callable[[A], bool]) -> Iterator[A]:
    """
    Selects all elements of this set which do not satisfy a predicate, lazily.

    Args:
        p: The predicate to not satisfy.

    Returns:
        The filtered lazy set, as an iterator.
    """
    for x in self:
        if not p(x):
            yield x


def lazy_flat_map(self: Set[A], f: Callable[[A], Set[B]]) -> Iterator[B]:
    """
    Builds a new lazy set by applying a function to all elements of this set and using the elements of the resulting collections.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new lazy set, as an iterator.
    """
    return (y for x in self for y in f(x))



from warnings import warn

from . import dict as pfcdict
from . import list as pfclist
from . import set as pfcset

EXTEND_BUILTINS = False
try:
    from forbiddenfruit import curse

    EXTEND_BUILTINS = True
except ImportError:
    warn(
        "[WARNING] You are using pyfuncol without forbiddenfruit. Functions will not extend builtins"
    )


def extend_dict():
    """
    Extends the dict built-in type with methods.
    """
    curse(dict, "contains", pfcdict.contains)
    curse(dict, "size", pfcdict.size)
    curse(dict, "filter", pfcdict.filter)
    curse(dict, "filter_not", pfcdict.filter_not)
    curse(dict, "flat_map", pfcdict.flat_map)
    curse(dict, "foreach", pfcdict.foreach)
    curse(dict, "is_empty", pfcdict.is_empty)
    curse(dict, "map", pfcdict.map)
    curse(dict, "to_list", pfcdict.to_list)
    curse(dict, "to_iterator", pfcdict.to_iterator)
    curse(dict, "count", pfcdict.count)
    curse(dict, "fold_left", pfcdict.fold_left)
    curse(dict, "fold_right", pfcdict.fold_right)
    curse(dict, "forall", pfcdict.forall)
    curse(dict, "find", pfcdict.find)

    # Parallel operations
    curse(dict, "par_map", pfcdict.par_map)
    curse(dict, "par_filter", pfcdict.par_filter)
    curse(dict, "par_filter_not", pfcdict.par_filter_not)
    curse(dict, "par_flat_map", pfcdict.par_flat_map)

    # Pure operations
    curse(dict, "pure_map", pfcdict.pure_map)
    curse(dict, "pure_flat_map", pfcdict.pure_flat_map)
    curse(dict, "pure_filter", pfcdict.pure_filter)
    curse(dict, "pure_filter_not", pfcdict.pure_filter_not)

    # Lazy operations
    curse(dict, "lazy_map", pfcdict.lazy_map)
    curse(dict, "lazy_flat_map", pfcdict.lazy_flat_map)
    curse(dict, "lazy_filter", pfcdict.lazy_filter)
    curse(dict, "lazy_filter_not", pfcdict.lazy_filter_not)


def extend_list():
    """
    Extends the list built-in type with methods.
    """
    curse(list, "map", pfclist.map)
    curse(list, "filter", pfclist.filter)
    curse(list, "filter_not", pfclist.filter_not)
    curse(list, "flat_map", pfclist.flat_map)
    curse(list, "flatten", pfclist.flatten)
    curse(list, "contains", pfclist.contains)
    curse(list, "distinct", pfclist.distinct)
    curse(list, "foreach", pfclist.foreach)
    curse(list, "group_by", pfclist.group_by)
    curse(list, "is_empty", pfclist.is_empty)
    curse(list, "size", pfclist.size)
    curse(list, "find", pfclist.find)
    curse(list, "index_of", pfclist.index_of)
    curse(list, "fold_left", pfclist.fold_left)
    curse(list, "fold_right", pfclist.fold_right)
    curse(list, "forall", pfclist.forall)
    curse(list, "head", pfclist.head)
    curse(list, "tail", pfclist.tail)
    curse(list, "take", pfclist.take)
    curse(list, "length", pfclist.length)
    curse(list, "to_iterator", pfclist.to_iterator)

    # Parallel operations
    curse(list, "par_map", pfclist.par_map)
    curse(list, "par_filter", pfclist.par_filter)
    curse(list, "par_filter_not", pfclist.par_filter_not)
    curse(list, "par_flat_map", pfclist.par_flat_map)

    # Pure operations
    curse(list, "pure_map", pfclist.pure_map)
    curse(list, "pure_flat_map", pfclist.pure_flat_map)
    curse(list, "pure_filter", pfclist.pure_filter)
    curse(list, "pure_filter_not", pfclist.pure_filter_not)

    # Lazy operations
    curse(list, "lazy_map", pfclist.lazy_map)
    curse(list, "lazy_flat_map", pfclist.lazy_flat_map)
    curse(list, "lazy_filter", pfclist.lazy_filter)
    curse(list, "lazy_filter_not", pfclist.lazy_filter_not)
    curse(list, "lazy_flatten", pfclist.lazy_flatten)
    curse(list, "lazy_distinct", pfclist.lazy_distinct)
    curse(list, "lazy_take", pfclist.lazy_take)


def extend_set():
    """
    Extends the set and frozenset built-in type with methods.
    """
    curse(set, "map", pfcset.map)
    curse(set, "filter", pfcset.filter)
    curse(set, "filter_not", pfcset.filter_not)
    curse(set, "flat_map", pfcset.flat_map)
    curse(set, "contains", pfcset.contains)
    curse(set, "foreach", pfcset.foreach)
    curse(set, "group_by", pfcset.group_by)
    curse(set, "is_empty", pfcset.is_empty)
    curse(set, "size", pfcset.size)
    curse(set, "find", pfcset.find)
    curse(set, "fold_left", pfcset.fold_left)
    curse(set, "fold_right", pfcset.fold_right)
    curse(set, "forall", pfcset.forall)
    curse(set, "length", pfcset.length)
    curse(set, "to_iterator", pfcset.to_iterator)

    curse(frozenset, "map", pfcset.map)
    curse(frozenset, "filter", pfcset.filter)
    curse(frozenset, "filter_not", pfcset.filter_not)
    curse(frozenset, "flat_map", pfcset.flat_map)
    curse(frozenset, "contains", pfcset.contains)
    curse(frozenset, "foreach", pfcset.foreach)
    curse(frozenset, "group_by", pfcset.group_by)
    curse(frozenset, "is_empty", pfcset.is_empty)
    curse(frozenset, "size", pfcset.size)
    curse(frozenset, "find", pfcset.find)
    curse(frozenset, "fold_left", pfcset.fold_left)
    curse(frozenset, "fold_right", pfcset.fold_right)
    curse(frozenset, "forall", pfcset.forall)
    curse(frozenset, "length", pfcset.length)
    curse(frozenset, "to_iterator", pfcset.to_iterator)

    # Parallel operations
    curse(set, "par_map", pfcset.par_map)
    curse(set, "par_filter", pfcset.par_filter)
    curse(set, "par_filter_not", pfcset.par_filter_not)
    curse(set, "par_flat_map", pfcset.par_flat_map)

    curse(frozenset, "par_map", pfcset.par_map)
    curse(frozenset, "par_filter", pfcset.par_filter)
    curse(frozenset, "par_filter_not", pfcset.par_filter_not)
    curse(frozenset, "par_flat_map", pfcset.par_flat_map)

    # Pure operations
    curse(set, "pure_map", pfcset.pure_map)
    curse(set, "pure_flat_map", pfcset.pure_flat_map)
    curse(set, "pure_filter", pfcset.pure_filter)
    curse(set, "pure_filter_not", pfcset.pure_filter_not)

    curse(frozenset, "pure_map", pfcset.pure_map)
    curse(frozenset, "pure_flat_map", pfcset.pure_flat_map)
    curse(frozenset, "pure_filter", pfcset.pure_filter)
    curse(frozenset, "pure_filter_not", pfcset.pure_filter_not)

    # Lazy operations
    curse(set, "lazy_map", pfcset.lazy_map)
    curse(set, "lazy_flat_map", pfcset.lazy_flat_map)
    curse(set, "lazy_filter", pfcset.lazy_filter)
    curse(set, "lazy_filter_not", pfcset.lazy_filter_not)

    curse(frozenset, "lazy_map", pfcset.lazy_map)
    curse(frozenset, "lazy_flat_map", pfcset.lazy_flat_map)
    curse(frozenset, "lazy_filter", pfcset.lazy_filter)
    curse(frozenset, "lazy_filter_not", pfcset.lazy_filter_not)


if EXTEND_BUILTINS:
    extend_dict()
    extend_list()
    extend_set()



from typing import Callable, Dict, Iterator, Optional, Tuple, TypeVar, List, cast
import functools
import dask

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")
D = TypeVar("D")
U = TypeVar("U")


def contains(self: Dict[A, B], key: A) -> bool:
    """
    Tests whether this dict contains a binding for a key.

    Args:
        key: The key to find.

    Returns:
        True if the dict contains a binding for the key, False otherwise.
    """
    return key in self


def size(self: Dict[A, B]) -> int:
    """
    Computes the size of this dict.

    Returns:
        The size of the dict.
    """
    return len(self)


def filter(self: Dict[A, B], p: Callable[[Tuple[A, B]], bool]) -> Dict[A, B]:
    """
    Selects all elements of this dict which satisfy a predicate.

    Args:
        p: The predicate to satisfy.

    Returns:
        The filtered dict.
    """
    return type(self)({k: v for k, v in self.items() if p((k, v))})


def filter_not(self: Dict[A, B], p: Callable[[Tuple[A, B]], bool]) -> Dict[A, B]:
    """
    Selects all elements of this dict which do not satisfy a predicate.

    Args:
        p: The predicate to not satisfy.

    Returns:
        The filtered dict.
    """
    return type(self)({k: v for k, v in self.items() if not p((k, v))})


def flat_map(self: Dict[A, B], f: Callable[[Tuple[A, B]], Dict[C, D]]) -> Dict[C, D]:
    """
    Builds a new dict by applying a function to all elements of this dict and using the elements of the resulting collections.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new dict.
    """
    res = cast(Dict[C, D], type(self)())
    for k, v in self.items():
        d = f((k, v))
        res.update(d)
    return res


def foreach(self: Dict[A, B], f: Callable[[Tuple[A, B]], U]) -> None:
    """
    Apply f to each element for its side effects.

    Args:
        f: The function to apply to all elements for its side effects.
    """
    for k, v in self.items():
        f((k, v))


def is_empty(self: Dict[A, B]) -> bool:
    """
    Tests whether the dict is empty.

    Returns:
        True if the dict is empty, False otherwise.
    """
    return len(self) == 0


def map(self: Dict[A, B], f: Callable[[Tuple[A, B]], Tuple[C, D]]) -> Dict[C, D]:
    """
    Builds a new dict by applying a function to all elements of this dict.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new dict.
    """
    return cast(Dict[C, D], type(self)(f(x) for x in self.items()))


def to_list(self: Dict[A, B]) -> List[Tuple[A, B]]:
    """
    Converts this dict to a list of (key, value) pairs.

    Returns:
        A list of pairs corresponding to the entries of the dict
    """
    return [(k, v) for k, v in self.items()]


def to_iterator(self: Dict[A, B]) -> Iterator[Tuple[A, B]]:
    """
    Converts this dict to an iterator of (key, value) pairs.

    Returns:
        An iterator of pairs corresponding to the entries of the dict
    """
    return ((k, v) for k, v in self.items())


def count(self: Dict[A, B], p: Callable[[Tuple[A, B]], bool]) -> int:
    """
    Counts the number of elements in the collection which satisfy a predicate.

    Note: will not terminate for infinite-sized collections.

    Args:
        p: The predicate used to test elements.

    Returns:
        The number of elements satisfying the predicate p.
    """
    c = 0
    for t in self.items():
        if p(t):
            c += 1

    return c


def fold_left(self: Dict[A, B], z: B, op: Callable[[B, Tuple[A, B]], B]) -> B:
    """
    Applies a binary operator to a start value and all elements of this collection, going left to right.

    Note: will not terminate for infinite-sized collections.

    Note: might return different results for different runs, unless the underlying collection type is ordered or the operator is associative and commutative.

    Args:
        z: The start value.
        op: The binary operator.

    Returns:
        The result of inserting op between consecutive elements of this collection, going left to right with the start value z on the left:

        op(...op(z, x_1), x_2, ..., x_n)
        where x1, ..., xn are the elements of this collection. Returns z if this collection is empty.
    """
    acc = z
    for t in self.items():
        acc = op(acc, t)

    return acc


def fold_right(self: Dict[A, B], z: B, op: Callable[[Tuple[A, B], B], B]) -> B:
    """
    Applies a binary operator to a start value and all elements of this collection, going right to left.

    Note: will not terminate for infinite-sized collections.

    Note: might return different results for different runs, unless the underlying collection type is ordered or the operator is associative and commutative.

    Args:
        z: The start value.
        op: The binary operator.

    Returns:
        The result of inserting op between consecutive elements of this collection, going right to left with the start value z on the right:

        op(x_1, op(x_2, ... op(x_n, z)...))
        where x1, ..., xn are the elements of this collection. Returns z if this collection is empty.
    """
    acc = z
    for t in reversed(self.items()):
        acc = op(t, acc)

    return acc


def forall(self: Dict[A, B], p: Callable[[Tuple[A, B]], bool]) -> bool:
    """
    Tests whether a predicate holds for all elements of this collection.

    Note: may not terminate for infinite-sized collections.

    Args:
        p: The predicate used to test elements.

    Returns:
        True if this collection is empty or the given predicate p holds for all elements of this collection, otherwise False.
    """
    for t in self.items():
        if not p(t):
            return False
    return True


def find(self: Dict[A, B], p: Callable[[Tuple[A, B]], bool]) -> Optional[Tuple[A, B]]:
    """
    Finds the first element of the collection satisfying a predicate, if any.

    Note: may not terminate for infinite-sized collections.

    Note: might return different results for different runs, unless the underlying collection type is ordered.

    Args:
        p: The predicate used to test elements.

    Returns:
        An option value containing the first element in the collection that satisfies p, or None if none exists.
    """
    for t in self.items():
        if p(t):
            return t

    return None


# Parallel operations


def par_filter(self: Dict[A, B], p: Callable[[Tuple[A, B]], bool]) -> Dict[A, B]:
    """
    Selects in parallel all elements of this dict which satisfy a predicate.

    Args:
        p: The predicate to satisfy.

    Returns:
        The filtered dict.
    """
    preds = dask.compute(*(dask.delayed(p)(x) for x in self.items()))
    return type(self)({k: v for i, (k, v) in enumerate(self.items()) if preds[i]})


def par_filter_not(self: Dict[A, B], p: Callable[[Tuple[A, B]], bool]) -> Dict[A, B]:
    """
    Selects in parallel all elements of this dict which do not satisfy a predicate.

    Args:
        p: The predicate to not satisfy.

    Returns:
        The filtered dict.
    """
    preds = dask.compute(*(dask.delayed(p)(x) for x in self.items()))
    return type(self)({k: v for i, (k, v) in enumerate(self.items()) if not preds[i]})


def par_flat_map(
    self: Dict[A, B], f: Callable[[Tuple[A, B]], Dict[C, D]]
) -> Dict[C, D]:
    """
    Builds a new dict by applying a function in parallel to all elements of this dict and using the elements of the resulting collections.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new dict.
    """
    applications = dask.compute(*(dask.delayed(f)(x) for x in self.items()))
    return cast(
        Dict[C, D], type(self)({k: v for y in applications for k, v in y.items()})
    )


def par_map(self: Dict[A, B], f: Callable[[Tuple[A, B]], Tuple[C, D]]) -> Dict[C, D]:
    """
    Builds a new dict by applying a function in parallel to all elements of this dict.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new dict.
    """
    return cast(
        Dict[C, D],
        type(self)((dask.compute(*(dask.delayed(f)(x) for x in self.items())))),
    )


# Pure operations


def pure_map(self: Dict[A, B], f: Callable[[Tuple[A, B]], Tuple[C, D]]) -> Dict[C, D]:
    """
    Builds a new dict by applying a function to all elements of this dict using memoization to improve performance.

    WARNING: f must be a PURE function i.e., calling f on the same input must always lead to the same result!

    Type A must be hashable using `hash()` function.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new dict.
    """
    f_cache = functools.cache(f)
    return cast(Dict[C, D], type(self)(f_cache(x) for x in self.items()))


def pure_flat_map(
    self: Dict[A, B], f: Callable[[Tuple[A, B]], Dict[C, D]]
) -> Dict[C, D]:
    """
    Builds a new dict by applying a function to all elements of this dict and using the elements of the resulting collections using memoization to improve performance.

    WARNING: f must be a PURE function i.e., calling f on the same input must always lead to the same result!

    Type A must be hashable using `hash()` function.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new dict.
    """
    res = cast(Dict[C, D], type(self)())
    f_cache = functools.cache(f)
    for k, v in self.items():
        d = f_cache((k, v))
        res.update(d)
    return res


def pure_filter(self: Dict[A, B], p: Callable[[Tuple[A, B]], bool]) -> Dict[A, B]:
    """
    Selects all elements of this dict which satisfy a predicate using memoization to improve performance.

    WARNING: p must be a PURE function i.e., calling p on the same input must always lead to the same result!

    Type A must be hashable using `hash()` function.


    Args:
        p: The predicate to satisfy.

    Returns:
        The filtered dict.
    """
    p_cache = functools.cache(p)
    return type(self)({k: v for k, v in self.items() if p_cache((k, v))})


def pure_filter_not(self: Dict[A, B], p: Callable[[Tuple[A, B]], bool]) -> Dict[A, B]:
    """
    Selects all elements of this dict which do not satisfy a predicate using memoization to improve performance.

    WARNING: p must be a PURE function i.e., calling p on the same input must always lead to the same result!

    Type A must be hashable using `hash()` function.


    Args:
        p: The predicate not to satisfy.

    Returns:
        The filtered dict.
    """
    p_cache = functools.cache(p)
    return type(self)({k: v for k, v in self.items() if not p_cache((k, v))})


# Lazy operations


def lazy_map(
    self: Dict[A, B], f: Callable[[Tuple[A, B]], Tuple[C, D]]
) -> Iterator[Tuple[C, D]]:
    """
    Builds a new list of tuples by applying a function to all elements of this dict, lazily.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new lazy list of tuples, as an iterator.
    """
    for x in self.items():
        yield f(x)


def lazy_flat_map(
    self: Dict[A, B], f: Callable[[Tuple[A, B]], Dict[C, D]]
) -> Iterator[Tuple[C, D]]:
    """
    Builds a new list of tuples by applying a function to all elements of this dict and using the elements of the resulting collections, lazily.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new lazy list of tuples, as an iterator.
    """
    return (y for x in self.items() for y in f(x).items())


def lazy_filter(
    self: Dict[A, B], p: Callable[[Tuple[A, B]], bool]
) -> Iterator[Tuple[A, B]]:
    """
    Selects all elements of this dict which satisfy a predicate, lazily.

    Args:
        p: The predicate to satisfy.

    Returns:
        The filtered lazy list of tuples, as an iterator.
    """
    for x in self.items():
        if p(x):
            yield x


def lazy_filter_not(
    self: Dict[A, B], p: Callable[[Tuple[A, B]], bool]
) -> Iterator[Tuple[A, B]]:
    """
    Selects all elements of this dict which do not satisfy a predicate, lazily.

    Args:
        p: The predicate to not satisfy.

    Returns:
        The filtered lazy list of tuples, as an iterator.
    """
    for x in self.items():
        if not p(x):
            yield x



from collections import defaultdict
from typing import Callable, Dict, Iterator, Optional, TypeVar, List, cast
import functools
import dask

A = TypeVar("A")
B = TypeVar("B")
K = TypeVar("K")
U = TypeVar("U")


def map(self: List[A], f: Callable[[A], B]) -> List[B]:
    """
    Builds a new list by applying a function to all elements of this list.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new list.
    """
    return cast(List[B], type(self)(f(x) for x in self))


def filter(self: List[A], p: Callable[[A], bool]) -> List[A]:
    """
    Selects all elements of this list which satisfy a predicate.

    Args:
        p: The predicate to satisfy.

    Returns:
        The filtered list.
    """
    return type(self)(x for x in self if p(x))


def filter_not(self: List[A], p: Callable[[A], bool]) -> List[A]:
    """
    Selects all elements of this list which do not satisfy a predicate.

    Args:
        p: The predicate to not satisfy.

    Returns:
        The filtered list.
    """
    return type(self)(x for x in self if not p(x))


def flat_map(self: List[A], f: Callable[[A], List[B]]) -> List[B]:
    """
    Builds a new list by applying a function to all elements of this list and using the elements of the resulting collections.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new list.
    """
    return cast(List[B], type(self)(y for x in self for y in f(x)))


def flatten(self: List[A]) -> List[B]:
    """
    Converts this list of lists into a list formed by the elements of these lists.

    Returns:
        The flattened list.
    """
    return cast(List[B], type(self)(y for x in self for y in x))


def contains(self: List[A], elem: A) -> bool:
    """
    Tests whether this list contains a given value as element.

    Args:
        elem: The element to look for.

    Returns:
        True if the list contains the element, False otherwise.
    """
    return elem in self


def distinct(self: List[A]) -> List[A]:
    """
    Selects all the elements of this list ignoring the duplicates.

    Returns:
        The list without duplicates.
    """
    return type(self)((set(self)))


def foreach(self: List[A], f: Callable[[A], U]) -> None:
    """
    Apply f to each element for its side effects.

    Args:
        f: The function to apply to all elements for its side effects.
    """
    for x in self:
        f(x)


def group_by(self: List[A], f: Callable[[A], K]) -> Dict[K, List[A]]:
    """
    Partitions this list into a dict of lists according to some discriminator function.

    Args:
        f: The grouping function.

    Returns:
        A dictionary where elements are grouped according to the grouping function.
    """
    d = defaultdict(type(self))
    for x in self:
        k = f(x)
        d[k].append(x)
    return dict(d)


def is_empty(self: List[A]) -> bool:
    """
    Tests whether the list is empty.

    Returns:
        True if the list is empty, False otherwise.
    """
    return len(self) == 0


def size(self: List[A]) -> int:
    """
    Computes the size of this list.

    Returns:
        The size of the list.
    """
    return len(self)


def find(self: List[A], p: Callable[[A], bool]) -> Optional[A]:
    """
    Finds the first element of the list satisfying a predicate, if any.

    Args:
        p: The predicate to satisfy.

    Returns:
        The first element satisfying the predicate, otherwise None.
    """
    for x in self:
        if p(x):
            return x
    return None


def index_of(self: List[A], elem: A) -> int:
    """
    Finds index of first occurrence of some value in this list. Returns -1 if none exists.

    Args:
        elem: The element whose index is to find.

    Returns:
        The index of the first occurrence of the element, or -1 if it does not exists.
    """
    for i, x in enumerate(self):
        if x == elem:
            return i
    return -1


def fold_left(self: List[A], z: B, op: Callable[[B, A], B]) -> B:
    """
    Applies a binary operator to a start value and all elements of this sequence, going left to right.

    Args:
        z: The start value.
        op: The binary operation.

    Returns:
        The result of inserting op between consecutive elements of this sequence, going left to right with the start value z on the left:
        op(...op(z, x_1), x_2, ..., x_n)
        where x1, ..., xn are the elements of this sequence. Returns z if this sequence is empty.
    """
    acc = z
    for x in self:
        acc = op(acc, x)
    return acc


def fold_right(self: List[A], z: B, op: Callable[[A, B], B]) -> B:
    """
    Applies a binary operator to all elements of this list and a start value, going right to left.

    Args:
        z: The start value.
        op: The binary operation.

    Returns:
        The result of inserting op between consecutive elements of this list, going right to left with the start value z on the right:
        op(x_1, op(x_2, ... op(x_n, z)...))
        where x1, ..., xn are the elements of this list. Returns z if this list is empty.
    """

    acc = z
    for x in reversed(self):
        acc = op(x, acc)
    return acc


def forall(self: List[A], p: Callable[[A], bool]) -> bool:
    """
    Tests whether a predicate holds for all elements of this list.

    Args:
        p: The predicate used to test elements.

    Returns:
        True if this list is empty or the given predicate p holds for all elements of this list, otherwise False.
    """
    for x in self:
        if not p(x):
            return False

    return True


def head(self: List[A]) -> A:
    """
    Selects the first element of this iterable collection.

    Note: might return different results for different runs, unless the underlying collection type is ordered.

    Raises:
        IndexError: If the iterable collection is empty.
    """
    if not self:
        raise IndexError()
    return self[0]


def tail(self: List[A]) -> List[A]:
    """
    The rest of the collection without its first element.

    Raises:
        IndexError: If the iterable collection is empty.
    """
    if not self:
        raise IndexError()
    return self[1:]


def take(self: List[A], n: int) -> List[A]:
    """
    Selects the first n elements.

    Args:
        n: The number of elements to take from this list.

    Returns:
        A list consisting only of the first n elements of this list, or else the whole list, if it has less than n elements. If n is negative, returns an empty list.
    """

    if n < 0:
        return type(self)()
    if len(self) <= n:
        return self

    return self[0:n]


def length(self: List[A]) -> int:
    """
    Returns the length (number of elements) of the list. `size` is an alias for length.

    Returns:
        The length of the list
    """
    return len(self)


def to_iterator(self: List[A]) -> Iterator[A]:
    """
    Converts this list to an iterator.

    Returns:
        An iterator
    """
    return (x for x in self)


# Parallel operations


def par_map(self: List[A], f: Callable[[A], B]) -> List[B]:
    """
    Builds a new list by applying a function in parallel to all elements of this list.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new list.
    """
    return cast(
        List[B], type(self)((dask.compute(*(dask.delayed(f)(x) for x in self))))
    )


def par_filter(self: List[A], p: Callable[[A], bool]) -> List[A]:
    """
    Selects in parallel all elements of this list which satisfy a predicate.

    Args:
        p: The predicate to satisfy.

    Returns:
        The filtered list.
    """
    preds = dask.compute(*(dask.delayed(p)(x) for x in self))
    return type(self)(x for i, x in enumerate(self) if preds[i])


def par_filter_not(self: List[A], p: Callable[[A], bool]) -> List[A]:
    """
    Selects in parallel all elements of this list which do not satisfy a predicate.

    Args:
        p: The predicate to not satisfy.

    Returns:
        The filtered list.
    """
    preds = dask.compute(*(dask.delayed(p)(x) for x in self))
    return type(self)(x for i, x in enumerate(self) if not preds[i])


def par_flat_map(self: List[A], f: Callable[[A], List[B]]) -> List[B]:
    """
    Builds a new list by applying a function in parallel to all elements of this list and using the
    elements of the resulting collections.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new list.
    """
    applications = dask.compute(*(dask.delayed(f)(x) for x in self))
    return cast(List[B], type(self)(z for y in applications for z in y))


def pure_map(self: List[A], f: Callable[[A], B]) -> List[B]:
    """
    Builds a new list by applying a function to all elements of this list using memoization to improve performance.

    WARNING: f must be a PURE function i.e., calling f on the same input must always lead to the same result!

    Type A must be hashable using `hash()` function.

    Args:
        f: The PURE function to apply to all elements.

    Returns:
        The new list.
    """
    f_cache = functools.cache(f)
    return cast(List[B], type(self)(f_cache(x) for x in self))


def pure_flat_map(self: List[A], f: Callable[[A], List[B]]) -> List[B]:
    """
    Builds a new list by applying a function to all elements of this list and using the elements of the resulting collections using memoization to improve performance.

    WARNING: f must be a PURE function i.e., calling f on the same input must always lead to the same result!

    Type A must be hashable using `hash()` function.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new list.
    """
    f_cache = functools.cache(f)
    return cast(List[B], type(self)(y for x in self for y in f_cache(x)))


def pure_filter(self: List[A], p: Callable[[A], bool]) -> List[A]:
    """
    Selects all elements of this list which satisfy a predicate using memoization to improve performance.

    WARNING: p must be a PURE function i.e., calling p on the same input must always lead to the same result!

    Type A must be hashable using `hash()` function.

    Args:
        p: The predicate to satisfy.

    Returns:
        The filtered list.
    """
    p_cache = functools.cache(p)
    return type(self)(x for x in self if p_cache(x))


def pure_filter_not(self: List[A], p: Callable[[A], bool]) -> List[A]:
    """
    Selects all elements of this list which do not satisfy a predicate using memoization to improve performance.

    WARNING: p must be a PURE function i.e., calling p on the same input must always lead to the same result!

    Type A must be hashable using `hash()` function.


    Args:
        p: The predicate not to satisfy.

    Returns:
        The filtered list.
    """
    p_cache = functools.cache(p)
    return type(self)(x for x in self if not p_cache(x))


def lazy_map(self: List[A], f: Callable[[A], B]) -> Iterator[B]:
    """
    Builds a new list by applying a function to all elements of this list, lazily.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new lazy list, as an iterator.
    """
    for x in self:
        yield f(x)


def lazy_filter(self: List[A], p: Callable[[A], bool]) -> Iterator[A]:
    """
    Selects all elements of this list which satisfy a predicate, lazily.

    Args:
        p: The predicate to satisfy.

    Returns:
        The filtered lazy list, as an iterator.
    """
    for x in self:
        if p(x):
            yield x


def lazy_filter_not(self: List[A], p: Callable[[A], bool]) -> Iterator[A]:
    """
    Selects all elements of this list which do not satisfy a predicate, lazily.

    Args:
        p: The predicate to not satisfy.

    Returns:
        The filtered lazy list, as an iterator.
    """
    for x in self:
        if not p(x):
            yield x


def lazy_flat_map(self: List[A], f: Callable[[A], List[B]]) -> Iterator[B]:
    """
    Builds a new lazy list by applying a function to all elements of this list and using the elements of the resulting collections.

    Args:
        f: The function to apply to all elements.

    Returns:
        The new lazy list, as an iterator.
    """
    return (y for x in self for y in f(x))


def lazy_flatten(self: List[A]) -> Iterator[B]:
    """
    Converts this list of lists into a lazy list formed by the elements of these lists.

    Returns:
        The flattened lazy list, as an iterator.
    """
    return (y for x in self for y in x)


def lazy_distinct(self: List[A]) -> Iterator[A]:
    """
    Selects all the elements of this list ignoring the duplicates, lazily.

    Returns:
        The lazy list without duplicates, as an iterator.
    """
    for x in set(self):
        yield x


def lazy_take(self: List[A], n: int) -> Iterator[A]:
    """
    Selects the first n elements, lazily.

    Args:
        n: The number of elements to take from this list.

    Returns:
        A lazy list (as an iterator) consisting only of the first n elements of this list,
        or else the whole lazy list, if it has less than n elements.
        If n is negative, returns an empty lazy list.
    """
    if n < 0:
        yield from ()
    if len(self) <= n:
        for x in self:
            yield x
    else:
        for i in range(n):
            yield self[i]



from setuptools import setup, find_packages

setup(
    name="pyfuncol",
    description="Functional collections extension functions for Python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/didactic-meme/pyfuncol",
    author="Andrea Veneziano",
    author_email="andrea.veneziano@icloud.com",
    maintainer="Andrea Veneziano",
    maintainer_email="andrea.veneziano@icloud.com",
    license="MIT",
    keywords="functional pipeline data collection chain parallel",
    packages=find_packages(exclude=["contrib", "docs", "*tests*", "test"]),
    version="1.3.0",
    install_requires=["forbiddenfruit", "dask"],
    extras_requires={"dev": ["pytest", "pytest-cov", "myst-parser", "black", "Sphinx", "sphinx-rtd-theme"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
